# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2009, Alessandro Decina <alessandro.decina@collabora.co.uk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import collections.abc
import errno
import fnmatch
import os
import re
import sys
import threading
import time
import traceback
import types


# environment variables controlling levels for each category
_DEBUG = "*:1"
# name of the environment variable controlling our logging
_ENV_VAR_NAME = None
# package names we should scrub filenames for
_PACKAGE_SCRUB_LIST = []

# dynamic dictionary of categories already seen and their level
_categories = {}

# log handlers registered
_log_handlers = []
_log_handlers_limited = []

_initialized = False
_enable_crack_output = False

_stdout = None
_stderr = None
_old_hup_handler = None
_outfile = None


# public log levels
(ERROR,
 WARN,
 FIXME,
 INFO,
 DEBUG,
 LOG) = list(range(1, 7))

COLORS = {ERROR: 'RED',
          WARN: 'YELLOW',
          FIXME: 'MAGENTA',
          INFO: 'GREEN',
          DEBUG: 'BLUE',
          LOG: 'CYAN'}

_FORMATTED_LEVELS = []
_LEVEL_NAMES = ['ERROR', 'WARN', 'FIXME', 'INFO', 'DEBUG', 'LOG']


class TerminalController:
    """A class for generating formatted output to a terminal.

    `TerminalController` defines a set of instance variables whose
    values are initialized to the control sequence necessary to
    perform a given action.  These can be simply included in normal
    output to the terminal:

        >>> term = TerminalController()
        >>> print('This is '+term.GREEN+'green'+term.NORMAL)

    Alternatively, the `render()` method can used, which replaces
    '${action}' with the string required to perform 'action':

        >>> term = TerminalController()
        >>> print(term.render('This is ${GREEN}green${NORMAL}'))

    If the terminal doesn't support a given action, then the value of
    the corresponding instance variable will be set to ''.  As a
    result, the above code will still work on terminals that do not
    support color, except that their output will not be colored.
    Also, this means that you can test whether the terminal supports a
    given action by simply testing the truth value of the
    corresponding instance variable:

        >>> term = TerminalController()
        >>> if term.CLEAR_SCREEN:
        ...     print('This terminal supports clearning the screen.')

    Finally, if the width and height of the terminal are known, then
    they will be stored in the `COLS` and `LINES` attributes.

    Args:
        term_stream (Optional): The stream that will be used for terminal
            output; if this stream is not a tty, then the terminal is
            assumed to be a dumb terminal (i.e., have no capabilities).
    """

    # Cursor movement:
    BOL = ''             # : Move the cursor to the beginning of the line
    UP = ''              # : Move the cursor up one line
    DOWN = ''            # : Move the cursor down one line
    LEFT = ''            # : Move the cursor left one char
    RIGHT = ''           # : Move the cursor right one char

    # Deletion:
    CLEAR_SCREEN = ''    # : Clear the screen and move to home position
    CLEAR_EOL = ''       # : Clear to the end of the line.
    CLEAR_BOL = ''       # : Clear to the beginning of the line.
    CLEAR_EOS = ''       # : Clear to the end of the screen

    # Output modes:
    BOLD = ''            # : Turn on bold mode
    BLINK = ''           # : Turn on blink mode
    DIM = ''             # : Turn on half-bright mode
    REVERSE = ''         # : Turn on reverse-video mode
    NORMAL = ''          # : Turn off all modes

    # Cursor display:
    HIDE_CURSOR = ''     # : Make the cursor invisible
    SHOW_CURSOR = ''     # : Make the cursor visible

    # Terminal size:
    COLS = None          # : Width of the terminal (None for unknown)
    LINES = None         # : Height of the terminal (None for unknown)

    # Foreground colors:
    BLACK = BLUE = GREEN = CYAN = RED = MAGENTA = YELLOW = WHITE = ''

    # Background colors:
    BG_BLACK = BG_BLUE = BG_GREEN = BG_CYAN = ''
    BG_RED = BG_MAGENTA = BG_YELLOW = BG_WHITE = ''

    _STRING_CAPABILITIES = """
    BOL=cr UP=cuu1 DOWN=cud1 LEFT=cub1 RIGHT=cuf1
    CLEAR_SCREEN=clear CLEAR_EOL=el CLEAR_BOL=el1 CLEAR_EOS=ed BOLD=bold
    BLINK=blink DIM=dim REVERSE=rev UNDERLINE=smul NORMAL=sgr0
    HIDE_CURSOR=cinvis SHOW_CURSOR=cnorm""".split()
    _COLORS = """BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE""".split()
    _ANSICOLORS = "BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE".split()

    def __init__(self, term_stream=sys.stdout):
        # Curses isn't available on all platforms
        try:
            import curses
        except ImportError:
            return

        # If the stream isn't a tty, then assume it has no capabilities.
        if not term_stream.isatty():
            return

        # Check the terminal type.  If we fail, then assume that the
        # terminal has no capabilities.
        try:
            curses.setupterm()
        except curses.error:
            return

        # Look up numeric capabilities.
        TerminalController.COLS = curses.tigetnum('cols')
        TerminalController.LINES = curses.tigetnum('lines')

        # Look up string capabilities.
        for capability in self._STRING_CAPABILITIES:
            (attrib, cap_name) = capability.split('=')
            setattr(self, attrib, self._tigetstr(cap_name) or b'')

        # Colors
        set_fg = self._tigetstr('setf')
        if set_fg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, color, curses.tparm(set_fg, i) or b'')
        set_fg_ansi = self._tigetstr('setaf')
        if set_fg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))),
                                self._ANSICOLORS):
                setattr(self, color, curses.tparm(set_fg_ansi, i) or b'')
        set_bg = self._tigetstr('setb')
        if set_bg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, 'BG_' + color, curses.tparm(set_bg, i) or b'')
        set_bg_ansi = self._tigetstr('setab')
        if set_bg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))),
                                self._ANSICOLORS):
                setattr(
                    self, 'BG_' + color, curses.tparm(set_bg_ansi, i) or b'')

    def _tigetstr(self, cap_name):
        # String capabilities can include "delays" of the form "$<2>".
        # For any modern terminal, we should be able to just ignore
        # these, so strip them out.
        import curses
        cap = curses.tigetstr(cap_name) or b''
        return re.sub(r'\$<\d+>[/*]?', '', cap.decode()).encode()

    def render(self, template):
        """Replaces each $-substitutions in the specified template string.

        The placeholders are replaced with the corresponding terminal control
        string (if it's defined) or '' (if it's not).
        """
        return re.sub(r'\$\$|\${\w+}', self._render_sub, template)

    def _render_sub(self, match):
        matched_group = match.group()
        if matched_group == '$$':
            return matched_group
        else:
            return getattr(self, matched_group[2:-1])


def get_level_name(level):
    """Returns the name of the specified log level.

    Args:
        level (int): The level we want to know the name.

    Returns:
        str: The name of the level.
    """
    assert isinstance(level, int) and 0 < level <= len(_LEVEL_NAMES), \
        TypeError("Bad debug level")
    return _LEVEL_NAMES[level - 1]


def get_level_int(level_name):
    """Returns the integer value of the levelName.

    Args:
        level_name (str): The string value of the level name.

    Returns:
        int: The value of the level name we are interested in.
    """
    assert isinstance(level_name, str) and level_name in _LEVEL_NAMES, \
        "Bad debug level name"
    return _LEVEL_NAMES.index(level_name) + 1


def get_formatted_level_name(level):
    assert isinstance(level, int) and 0 < level <= len(_LEVEL_NAMES), \
        TypeError("Bad debug level")
    return _FORMATTED_LEVELS[level - 1]


def register_category(category):
    """Registers the specified category in the debug system.

    A level will be assigned to it based on previous calls to setDebug.
    """
    # parse what level it is set to based on _DEBUG
    # example: *:2,admin:4
    global _DEBUG
    global _categories

    level = 0
    chunks = _DEBUG.split(',')
    for chunk in chunks:
        if not chunk:
            continue
        if ':' in chunk:
            spec, value = chunk.split(':')
        else:
            spec = '*'
            value = chunk

        # our glob is unix filename style globbing, so cheat with fnmatch
        # fnmatch.fnmatch didn't work for this, so don't use it
        if category in fnmatch.filter((category, ), spec):
            # we have a match, so set level based on string or int
            if not value:
                continue
            try:
                level = int(value)
            except ValueError:  # e.g. *; we default to most
                level = 5
    # store it
    _categories[category] = level


def get_category_level(category):
    """Gets the debug level at which the specified category is being logged.

    Registers the category and thus assigns a log level if it wasn't registered
    yet.

    Args:
        category (string): The category we are interested in.
    """
    global _categories
    if category not in _categories:
        register_category(category)
    return _categories[category]


def set_log_settings(state):
    """Updates the current log settings.

    This can restore an old saved log settings object returned by
    getLogSettings.

    Args:
        state: The settings to set.
    """
    global _DEBUG
    global _log_handlers
    global _log_handlers_limited

    (_DEBUG,
     _categories,
     _log_handlers,
     _log_handlers_limited) = state

    for category in _categories:
        register_category(category)


def get_log_settings():
    """Fetches the current log settings.

    The returned object can be sent to setLogSettings to restore the
    returned settings

    Returns:
        The current settings.
    """
    return (_DEBUG,
            _categories,
            _log_handlers,
            _log_handlers_limited)


def _can_shortcut_logging(category, level):
    if _log_handlers:
        # we have some loggers operating without filters, have to do
        # everything
        return False
    else:
        return level > get_category_level(category)


def scrub_filename(filename):
    """Scrubs the filename to a relative path."""
    global _PACKAGE_SCRUB_LIST
    for package in _PACKAGE_SCRUB_LIST:
        i = filename.rfind(package)
        if i > -1:
            return filename[i:]

    return filename


def get_file_line(where=-1):
    """Returns the filename and line number for the specified location.

    Args:
        where(int or function): If it's a (negative) integer, looks for
            the code entry in the current stack that is the given number
            of frames above this module.
            If it's a function, look for the code entry of the function.

    Returns:
        str, int, str: file, line, function_name.
    """
    co = None
    lineno = None
    name = None

    if isinstance(where, types.FunctionType):
        co = where.__code__
        lineno = co.co_firstlineno
        name = co.co_name
    elif isinstance(where, types.MethodType):
        co = where.__func__.__code__
        lineno = co.co_firstlineno
        name = co.co_name
    else:
        stack_frame = sys._getframe()  # pylint: disable=protected-access
        while stack_frame:
            co = stack_frame.f_code
            if not co.co_filename.endswith('loggable.py'):
                co = stack_frame.f_code
                lineno = stack_frame.f_lineno
                name = co.co_name
                break
            stack_frame = stack_frame.f_back

    if not co:
        return "<unknown file>", 0, None

    return scrub_filename(co.co_filename), lineno, name


def ellipsize(obj):
    """Ellipsizes the representation of the given object."""
    obj_repr = repr(obj)
    if len(obj_repr) < 800:
        return obj_repr

    obj_repr = obj_repr[:60] + ' ... ' + obj_repr[-15:]
    return obj_repr


def get_format_args(start_format, start_args, end_format, end_args, args, kwargs):
    """Creates a format and args to use for logging.

    This avoids needlessly interpolating variables.
    """
    debug_args = start_args[:]
    for arg in args:
        debug_args.append(ellipsize(arg))

    for items in list(kwargs.items()):
        debug_args.extend(items)
    debug_args.extend(end_args)
    fmt = start_format \
        + ', '.join(('%s', ) * len(args)) \
        + (kwargs and ', ' or '') \
        + ', '.join(('%s=%r', ) * len(kwargs)) \
        + end_format
    return fmt, debug_args


def do_log(level, obj, category, message, args, where=-1, file_path=None, line=None):
    """Logs something.

    Args:
        level (int): Debug level.
        obj (str): Object converted to str.
        category (str): Category such as the name of the obj's class.
        message (str): The message to log.
        args (list): The args to apply to the message, if any.
        where (int or function): What to log file and line number for;
            -1 for one frame above log.py; -2 and down for higher up;
            a function for a (future) code object.
        file_path (Optional[str]): The file to show the message as coming from,
            if caller knows best.
        line (Optional[int]): The line to show the message as coming from,
            if caller knows best.

    Returns:
        A dict of calculated variables, if they needed calculating.
        currently contains file and line; this prevents us from
        doing this work in the caller when it isn't needed because
        of the debug level.
    """
    ret = {}

    if args:
        message = message % args
    funcname = None

    if level > get_category_level(category):
        handlers = _log_handlers
    else:
        handlers = _log_handlers + _log_handlers_limited

    if handlers:
        if file_path is None and line is None:
            (file_path, line, funcname) = get_file_line(where=where)
        ret['filePath'] = file_path
        ret['line'] = line
        if funcname:
            message = "\033[00m\033[32;01m%s:\033[00m %s" % (funcname, message)
        for handler in handlers:
            try:
                handler(level, obj, category, file_path, line, message)
            except TypeError as e:
                raise SystemError("handler %r unusable" % handler) from e

    return ret


def error_object(obj, cat, fmt, *args):
    """Logs a fatal error message in the specified category.

    This will also raise a `SystemExit`.
    """
    do_log(ERROR, obj, cat, fmt, args)


def warning_object(obj, cat, fmt, *args):
    """Logs a warning message in the specified category.

    This is used for non-fatal problems.
    """
    do_log(WARN, obj, cat, fmt, args)


def fixme_object(obj, cat, fmt, *args):
    """Logs a fixme message in the specified category.

    This is used for not implemented codepaths or known issues in the code.
    """
    do_log(FIXME, obj, cat, fmt, args)


def info_object(obj, cat, fmt, *args):
    """Logs an informational message in the specified category."""
    do_log(INFO, obj, cat, fmt, args)


def debug_object(obj, cat, fmt, *args):
    """Logs a debug message in the specified category."""
    do_log(DEBUG, obj, cat, fmt, args)


def log_object(obj, cat, fmt, *args):
    """Logs a log message.

    Used for debugging recurring events.
    """
    do_log(LOG, obj, cat, fmt, args)


def safeprintf(file, fmt, *args):
    """Writes to a file object, ignoring errors."""
    try:
        if args:
            file.write(fmt % args)
        else:
            file.write(fmt)
    except IOError as e:
        if e.errno == errno.EPIPE:
            # if our output is closed, exit; e.g. when logging over an
            # ssh connection and the ssh connection is closed
            # pylint: disable=protected-access
            os._exit(os.EX_OSERR)
        # otherwise ignore it, there's nothing you can do


def print_handler(level, obj, category, file, line, message):
    """Writes to stderr.

    The output will be different depending the value of "_enable_crack_output";
    in Pitivi's case, that is True when the GST_DEBUG env var is defined.

    Args:
        level (str): The debug level.
        obj (Optional[str]): The object the message is about, or None.
        category (str): Category such as the name of the obj's class.
        file (str): The source file where the message originates.
        line (int): The line number in the file where the message originates.
        message (str): The message to be logged.
    """
    global _outfile

    # Make the file path more compact for readability
    file = os.path.relpath(file)
    where = "(%s:%d)" % (file, line)

    # If GST_DEBUG is not set, we can assume only PITIVI_DEBUG is set, so don't
    # show a bazillion of debug details that are not relevant to Pitivi.
    if not _enable_crack_output:
        safeprintf(_outfile, '%s %-8s %-17s %-2s %s %s\n',
                   get_formatted_level_name(level), time.strftime("%H:%M:%S"),
                   category, obj, message, where)
    else:
        if obj:
            obj = '"' + obj + '"'
        else:
            obj = ""
        # level   pid     object   cat      time
        # 5 + 1 + 7 + 1 + 32 + 1 + 17 + 1 + 15 == 80
        safeprintf(
            _outfile, '%s [%5d] [0x%12x] %-32s %-17s %-15s %-4s %s %s\n',
            get_formatted_level_name(level), os.getpid(),
            threading.current_thread().ident,
            obj[:32], category, time.strftime("%b %d %H:%M:%S"), "",
            message, where)
    _outfile.flush()


def log_level_name(level):
    fmt = '%-5s'
    return fmt % (_LEVEL_NAMES[level - 1], )


def _as_string(string_or_bytes):
    if isinstance(string_or_bytes, bytes):
        return string_or_bytes.decode()
    else:
        return string_or_bytes


def _preformat_levels(enable_color_output):
    terminal_controller = TerminalController()
    for level in ERROR, WARN, FIXME, INFO, DEBUG, LOG:
        if enable_color_output:
            formatter = ''.join(
                (_as_string(terminal_controller.BOLD),
                 _as_string(getattr(terminal_controller, COLORS[level])),
                 log_level_name(level),
                 _as_string(terminal_controller.NORMAL)))
        else:
            formatter = log_level_name(level)
        _FORMATTED_LEVELS.append(formatter)

# "public" useful API

# setup functions


def init(env_var_name, enable_color_output=True, enable_crack_output=True):
    """Initializes the logging system.

    Needs to be called before using the log methods.

    Args:
        env_var_name (str): The name of the environment variable with additional
            settings.
        enable_color_output (Optional[bool]): Whether to colorize the output.
        enable_crack_output (Optional[bool]): Whether to print detailed info.
    """
    global _initialized
    global _outfile
    global _enable_crack_output
    _enable_crack_output = enable_crack_output

    if _initialized:
        return

    global _ENV_VAR_NAME
    _ENV_VAR_NAME = env_var_name

    _preformat_levels(enable_color_output)

    if env_var_name in os.environ:
        # install a log handler that uses the value of the environment var
        set_debug(os.environ[env_var_name])
    filename_env_var_name = env_var_name + "_FILE"

    if filename_env_var_name in os.environ:
        # install a log handler that uses the value of the environment var
        # pylint: disable=consider-using-with
        _outfile = open(os.environ[filename_env_var_name], "w+", encoding="UTF-8")
    else:
        _outfile = sys.stderr

    add_limited_log_handler(print_handler)

    _initialized = True


def set_debug(string):
    """Sets the DEBUG string.

    This controls the log output.
    """
    global _DEBUG
    global _ENV_VAR_NAME
    global _categories

    _DEBUG = string
    debug('log', "%s set to %s" % (_ENV_VAR_NAME, _DEBUG))

    # reparse all already registered category levels
    for category in _categories:
        register_category(category)


def get_debug():
    """Returns the currently active DEBUG string."""
    global _DEBUG
    return _DEBUG


def set_package_scrub_list(*packages):
    """Sets the package names to scrub from filenames.

    Filenames from these paths in log messages will be scrubbed to their
    relative file path instead of the full absolute path.

    Args:
        *packages (List[str]): The packages names to scrub.
    """
    global _PACKAGE_SCRUB_LIST
    _PACKAGE_SCRUB_LIST = packages


def reset():
    """Resets the logging system, removing all log handlers."""
    global _log_handlers, _log_handlers_limited, _initialized

    _log_handlers = []
    _log_handlers_limited = []
    _initialized = False


def add_log_handler(func):
    """Adds a custom log handler.

    The log handler receives all the log messages.

    Args:
        func (function): A function object with prototype
            (level, obj, category, message) where level is either
            ERROR, WARN, INFO, DEBUG, or LOG, and the rest of the arguments are
            strings or None. Use getLevelName(level) to get a printable name
            for the log level.

    Raises:
        TypeError: When func is not a callable.
    """
    if not isinstance(func, collections.abc.Callable):
        raise TypeError("func must be callable")

    if func not in _log_handlers:
        _log_handlers.append(func)


def add_limited_log_handler(func):
    """Adds a custom limited log handler.

    The log handler receives only the messages passing the filter.

    Args:
        func (function): A function object with prototype
            (level, obj, category, message) where level is either
            ERROR, WARN, INFO, DEBUG, or LOG, and the rest of the arguments are
            strings or None. Use getLevelName(level) to get a printable name
            for the log level.

    Raises:
        TypeError: When func is not a callable.
    """
    if not isinstance(func, collections.abc.Callable):
        raise TypeError("func must be callable")

    if func not in _log_handlers_limited:
        _log_handlers_limited.append(func)


def remove_log_handler(func):
    """Removes a registered log handler.

    Raises:
        ValueError: When func is not registered.
    """
    _log_handlers.remove(func)


def remove_limited_log_handler(func):
    """Removes a registered limited log handler.

    Raises:
        ValueError: When func is not registered.
    """
    _log_handlers_limited.remove(func)

# public log functions


def error(cat, fmt, *args):
    error_object(None, cat, fmt, *args)


def warning(cat, fmt, *args):
    warning_object(None, cat, fmt, *args)


def fixme(cat, fmt, *args):
    fixme_object(None, cat, fmt, *args)


def info(cat, fmt, *args):
    info_object(None, cat, fmt, *args)


def debug(cat, fmt, *args):
    debug_object(None, cat, fmt, *args)


def log(cat, fmt, *args):
    log_object(None, cat, fmt, *args)

# public utility functions


def get_exception_message(exception, frame=-1, filename=None):
    """Returns a short message based on an exception.

    Useful for debugging.
    Tries to find where the exception was triggered.
    """
    stack = traceback.extract_tb(sys.exc_info()[2])
    if filename:
        stack = [f for f in stack if f[0].find(filename) > -1]
    # import code; code.interact(local=locals())
    (filename, line, func, text) = stack[frame]
    filename = scrub_filename(filename)
    exc = exception.__class__.__name__
    msg = ""
    # a shortcut to extract a useful message out of most exceptions
    # for now
    if str(exception):
        msg = ": %s" % str(exception)
    return "exception %(exc)s at %(filename)s:%(line)s: %(func)s()%(msg)s" \
        % locals()


# base class for loggable objects


class BaseLoggable:
    """Base class for objects that want to be able to log messages.

    The levels of severity for the messages are, in order from least
    to most: log, debug, info, warning, error.

    Attributes:
        log_category (str): The category under which the messages will be filed.
            Can be used to set a display filter.
    """

    def error(self, *args):
        """Logs an error.

        By default this will also raise an exception.
        """
        if _can_shortcut_logging(self.log_category, ERROR):
            return
        error_object(self.log_object_name(),
                     self.log_category, *self.log_function(*args))

    def warning(self, *args):
        """Logs a warning.

        Used for non-fatal problems.
        """
        if _can_shortcut_logging(self.log_category, WARN):
            return
        warning_object(
            self.log_object_name(), self.log_category, *self.log_function(*args))

    def fixme(self, *args):
        """Logs a fixme.

        Used for FIXMEs.
        """
        if _can_shortcut_logging(self.log_category, FIXME):
            return
        fixme_object(self.log_object_name(),
                     self.log_category, *self.log_function(*args))

    def info(self, *args):
        """Logs an informational message.

        Used for normal operation.
        """
        if _can_shortcut_logging(self.log_category, INFO):
            return
        info_object(self.log_object_name(),
                    self.log_category, *self.log_function(*args))

    def debug(self, *args):
        """Logs a debug message.

        Used for debugging.
        """
        if _can_shortcut_logging(self.log_category, DEBUG):
            return
        debug_object(self.log_object_name(),
                     self.log_category, *self.log_function(*args))

    def log(self, *args):
        """Logs a log message.

        Used for debugging recurring events.
        """
        if _can_shortcut_logging(self.log_category, LOG):
            return
        log_object(self.log_object_name(),
                   self.log_category, *self.log_function(*args))

    def do_log(self, level, where, fmt, *args, **kwargs):
        """Logs a message at the specified level.

        Args:
            level (int): The log level.
            where (int or function): How many frames to go back from
                the last log frame, must be negative; or a function
                (to log for a future call).
            fmt (str): The string template for the message.
            *args: The arguments used when converting the `fmt`
                string template to the message.
            **kwargs: The pre-calculated values from a previous do_log call.

        Returns:
            dict: The calculated variables, to be reused in a
                 call to do_log that should show the same location.
        """
        if _can_shortcut_logging(self.log_category, level):
            return {}
        args = self.log_function(*args)
        return do_log(level, self.log_object_name(), self.log_category,
                      fmt, args, where=where, **kwargs)

    def log_function(self, *args):
        """Processes the arguments applied to the message template.

        Default just returns the arguments unchanged.
        """
        return args

    def log_object_name(self):
        """Gets the name of this object."""
        for name in ['logName', 'name']:
            if hasattr(self, name):
                return getattr(self, name)

        return None

    def handle_exception(self, exc):
        self.warning(get_exception_message(exc))


class Loggable(BaseLoggable):

    def __init__(self, log_category=None):
        if log_category:
            self.log_category = log_category
        elif not hasattr(self, 'log_category'):
            self.log_category = self.__class__.__name__.lower()

    def log_object_name(self):
        res = BaseLoggable.log_object_name(self)
        if not res:
            return "<%s at 0x%x>" % (self.__class__.__name__, id(self))
        return res

    def error(self, fmt, *args):
        if _can_shortcut_logging(self.log_category, ERROR):
            return
        do_log(ERROR, self.log_object_name(), self.log_category,
               fmt, self.log_function(*args), where=-2)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2021, Piotr Brzeziński <thewildtreee@gmail.com>
# Copyright (C) 2022, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Widgets to control clips properties."""
import os
from gettext import gettext as _
from typing import Optional

from gi.repository import GES
from gi.repository import Gtk

from pitivi.check import MISSING_SOFT_DEPS
from pitivi.configure import get_ui_dir
from pitivi.utils.beat_detection import BeatDetector
from pitivi.utils.loggable import Loggable
from pitivi.utils.markers import GES_MARKERS_SNAPPABLE
from pitivi.utils.misc import disconnect_all_by_func
from pitivi.utils.ui import disable_scroll
from pitivi.utils.ui import SPACING


class ClipMarkersProperties(Gtk.Expander, Loggable):
    """Widget for managing the marker lists of a clip.

    Attributes:
        app (Pitivi): The app.
        clip (GES.Clip): The clip being configured.
    """

    TRACK_TYPES = {
        GES.TrackType.VIDEO: _("Video"),
        GES.TrackType.AUDIO: _("Audio"),
        GES.TrackType.TEXT: _("Text"),
        GES.TrackType.CUSTOM: _("Custom"),
    }

    def __init__(self, app):
        Gtk.Expander.__init__(self)
        Loggable.__init__(self)

        self.app = app
        self.clip: Optional[GES.Clip] = None
        self.beat_detector: Optional[BeatDetector] = None

        self._detect_button: Optional[Gtk.Button] = None
        self._clear_button: Optional[Gtk.Button] = None
        self._progress_bar: Optional[Gtk.ProgressBar] = None

        self.set_expanded(True)
        self.set_label(_("Clip markers"))

        self.expander_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.add(self.expander_box)

    def set_clip(self, clip):
        if self.clip:
            for child in self.clip.get_children(False):
                if not isinstance(child, GES.Source):
                    continue

                disconnect_all_by_func(child.markers_manager, self._lists_modified_cb)
                disconnect_all_by_func(child.markers_manager, self._current_list_changed_cb)

        for child in self.expander_box.get_children():
            self.expander_box.remove(child)

        self.clip = clip
        if not self.clip or not isinstance(self.clip, GES.SourceClip):
            self.hide()
            return

        self.show()

        audio_source: Optional[GES.AudioSource] = None
        labels_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        combos_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        for child in self.clip.get_children(False):
            # Ignore non-source children, e.g. effects
            if not isinstance(child, GES.Source):
                continue

            row_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=SPACING)
            row_box.set_border_width(SPACING)
            row_box.show()
            row_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.VERTICAL)

            child_type = child.get_track_type()
            name = ClipMarkersProperties.TRACK_TYPES[child_type]
            label = Gtk.Label(label=name)
            row_box.pack_start(label, False, False, 0)
            label.show()
            labels_size_group.add_widget(label)
            row_size_group.add_widget(label)
            label.props.valign = Gtk.Align.START

            controls_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=SPACING)
            controls_box.show()

            selection_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=SPACING)
            list_store = Gtk.ListStore(str, str)
            list_combo = Gtk.ComboBox.new_with_model(list_store)
            renderer_text = Gtk.CellRendererText()
            list_combo.pack_start(renderer_text, True)
            list_combo.add_attribute(renderer_text, "text", 1)
            list_combo.set_id_column(0)
            selection_box.pack_start(list_combo, False, False, 0)
            list_combo.show()
            combos_size_group.add_widget(list_combo)
            row_size_group.add_widget(list_combo)

            snap_toggle = Gtk.CheckButton.new_with_label(_("Magnetic"))
            selection_box.pack_start(snap_toggle, False, False, 0)
            if GES_MARKERS_SNAPPABLE:
                snap_toggle.show()

            controls_box.pack_start(selection_box, False, False, 0)

            if isinstance(child, GES.AudioSource) and "librosa" not in MISSING_SOFT_DEPS:
                container = self._create_beat_detection_ui()
                controls_box.pack_start(container, False, False, 0)
                audio_source = child

            row_box.pack_start(controls_box, True, True, 0)

            manager = child.markers_manager
            list_combo.connect("changed", self._combo_changed_cb, child, snap_toggle)
            snap_toggle.connect("toggled", self._snappable_toggled_cb, manager)

            self._populate_list_combo(manager, list_combo)
            manager.connect("lists-modified", self._lists_modified_cb, list_combo)
            manager.connect("current-list-changed", self._current_list_changed_cb, list_combo)

            selection_box.show()

            # Display audio marker settings below the video ones,
            # matching how they're shown on the timeline.
            if child_type == GES.TrackType.AUDIO:
                self.expander_box.pack_end(row_box, False, False, 0)
            else:
                self.expander_box.pack_start(row_box, False, False, 0)

        self._set_audio_source(audio_source)

        self.expander_box.show()
        disable_scroll(self.expander_box)

    def _current_list_changed_cb(self, manager, list_key, list_combo):
        list_combo.set_active_id(list_key)

    def _lists_modified_cb(self, manager, list_combo):
        self._populate_list_combo(manager, list_combo)

    def _populate_list_combo(self, manager, list_combo):
        lists = manager.get_all_keys_with_names()
        list_store = list_combo.get_model()

        list_store.clear()
        for key, name in lists:
            list_store.append([key, name])

        list_key = manager.current_list_key
        list_combo.set_active_id(list_key)

    def _combo_changed_cb(self, combo, ges_source, snap_toggle):
        tree_iter = combo.get_active_iter()
        if tree_iter is None:
            return

        model = combo.get_model()
        list_key = model[tree_iter][0]

        manager = ges_source.markers_manager
        manager.current_list_key = list_key

        snap_toggle.set_active(manager.snappable)
        snap_toggle_interactable = bool(list_key != "")
        snap_toggle.set_sensitive(snap_toggle_interactable)

    def _snappable_toggled_cb(self, button, manager):
        active = button.get_active()
        manager.snappable = active

    def _create_beat_detection_ui(self) -> Gtk.Box:
        builder = Gtk.Builder()
        builder.add_from_file(os.path.join(get_ui_dir(), "beatdetection.ui"))
        builder.connect_signals(self)

        self._detect_button = builder.get_object("detect-button")
        self._clear_button = builder.get_object("clear-button")
        self._progress_bar = builder.get_object("detection-progress")

        return builder.get_object("container-box")

    def _set_audio_source(self, source: GES.AudioSource):
        if self.beat_detector:
            self.beat_detector.disconnect_by_func(self._detection_percentage_cb)
            self.beat_detector.disconnect_by_func(self._detection_failed_cb)
            self.beat_detector = None

        if source:
            self.beat_detector = BeatDetector(source)
            self.beat_detector.connect("detection-percentage", self._detection_percentage_cb)
            self.beat_detector.connect("detection-failed", self._detection_failed_cb)

            self._update_beat_detection_ui(self.beat_detector.progress)

        self.set_visible(bool(source))

    def _detection_percentage_cb(self, detector, percentage):
        self._update_beat_detection_ui(percentage)

    def _detection_failed_cb(self, detector, error):
        # TODO: Show an error in the UI.
        self.error("BeatDetector failed: %s", error)
        self._update_beat_detection_ui()

    def _update_beat_detection_ui(self, percentage=0):
        in_progress = self.beat_detector.in_progress
        has_beats = self.beat_detector.beat_list_exists
        self._detect_button.set_sensitive(not in_progress and not has_beats)
        self._clear_button.set_sensitive(not in_progress and has_beats)

        self._progress_bar.set_visible(in_progress)
        self._progress_bar.set_fraction(percentage / 100)

    def _detect_clicked_cb(self, button):
        self.beat_detector.detect_beats()
        self._update_beat_detection_ui()

    def _clear_clicked_cb(self, button):
        self.beat_detector.clear_beats()
        self._update_beat_detection_ui()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import threading

from gi.repository import GObject

from pitivi.utils.loggable import Loggable

#
# Following code was freely adapted by code from:
#   John Stowers <john.stowers@gmail.com>
#


class Thread(threading.Thread, GObject.Object, Loggable):
    """Event-powered thread."""

    __gsignals__ = {
        "done": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        GObject.Object.__init__(self)
        threading.Thread.__init__(self)
        Loggable.__init__(self)

    def stop(self):
        """Stops the thread, do not override."""
        self.abort()
        self.emit("done")

    def run(self):
        """Runs the thread."""
        self.process()
        self.emit("done")

    def process(self):
        """Processes the thread.

        Implement this in subclasses.
        """
        raise NotImplementedError

    def abort(self):
        """Aborts the thread.

        Subclass have to implement this method !
        """


class ThreadMaster(Loggable):
    """Threads controller."""

    def __init__(self):
        Loggable.__init__(self)
        self.threads = []

    def add_thread(self, threadclass, *args):
        """Instantiates the specified Thread class and starts it."""
        assert issubclass(threadclass, Thread)
        self.log("Adding thread of type %r", threadclass)
        thread = threadclass(*args)
        thread.connect("done", self._thread_done_cb)
        self.threads.append(thread)
        self.log("starting it...")
        thread.start()
        self.log("started !")

    def _thread_done_cb(self, thread):
        self.log("thread %r is done", thread)
        self.threads.remove(thread)

    def wait_all_threads(self):
        """Waits until all running Threads controlled by this master stop."""
        self.log("Waiting for threads to stop")
        joinedthreads = 0
        while joinedthreads < len(self.threads):
            for thread in self.threads:
                self.log("Waiting for thread to stop: %r", thread)
                try:
                    thread.join()
                    joinedthreads += 1
                except RuntimeError:
                    # Tried to join the current thread, or one
                    # which did not start yet.
                    pass



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2012, Paul Lange <palango@gmx.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import re
from gettext import gettext as _

from gi.repository import Gdk
from gi.repository import GES
from gi.repository import Gio
from gi.repository import Gtk

from pitivi.timeline import elements
from pitivi.undo.timeline import CommitTimelineFinalizingAction
from pitivi.utils.loggable import Loggable
from pitivi.utils.timeline import Zoomable
from pitivi.utils.ui import LAYER_HEIGHT
from pitivi.utils.ui import PADDING
from pitivi.utils.ui import SEPARATOR_HEIGHT


AUDIO_ICONS = {
    True: "audio-volume-high-symbolic",
    False: "audio-volume-muted-symbolic",
}

VIDEO_ICONS = {
    True: "eye-open-negative-filled-symbolic",
    False: "eye-not-looking-symbolic",
}


class SpacedSeparator(Gtk.EventBox):
    """A Separator with vertical spacing.

    Inherits from EventBox since we want to change background color.
    """

    def __init__(self):
        Gtk.EventBox.__init__(self)

        self.get_style_context().add_class("SpacedSeparator")
        self.props.height_request = SEPARATOR_HEIGHT


class LayerControls(Gtk.EventBox, Loggable):
    """Container with widgets for controlling a layer."""

    __gtype_name__ = 'PitiviLayerControls'

    def __init__(self, ges_layer, app):
        Gtk.EventBox.__init__(self)
        Loggable.__init__(self)

        self.ges_layer = ges_layer
        self.ges_timeline = self.ges_layer.get_timeline()
        self.app = app
        self.__icon = None

        tracks = self.ges_timeline.get_tracks()
        self.timeline_audio_tracks = [track for track in tracks if track.props.track_type == GES.TrackType.AUDIO]
        self.timeline_video_tracks = [track for track in tracks if track.props.track_type == GES.TrackType.VIDEO]

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.add(hbox)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        hbox.pack_start(vbox, True, True, 0)

        rightside_separator = Gtk.Separator.new(Gtk.Orientation.VERTICAL)
        hbox.pack_start(rightside_separator, False, False, 0)

        name_row = Gtk.Box()
        name_row.set_orientation(Gtk.Orientation.HORIZONTAL)
        name_row.props.margin_top = PADDING
        name_row.props.margin_start = PADDING
        name_row.props.margin_end = PADDING
        vbox.pack_start(name_row, False, False, 0)

        self.menubutton = Gtk.MenuButton.new()
        self.menubutton.props.valign = Gtk.Align.CENTER
        self.menubutton.props.relief = Gtk.ReliefStyle.NONE
        model, action_group = self.__create_menu_model()
        popover = Gtk.Popover.new_from_model(self.menubutton, model)
        popover.insert_action_group("layer", action_group)
        popover.props.position = Gtk.PositionType.LEFT
        self.menubutton.set_popover(popover)
        name_row.pack_start(self.menubutton, False, False, 0)

        self.name_entry = Gtk.Entry()
        self.name_entry.get_style_context().add_class("LayerControlEntry")
        self.name_entry.props.valign = Gtk.Align.CENTER
        self.name_entry.connect("focus-out-event", self.__name_focus_out_cb)
        self.ges_layer.connect("notify-meta", self.__layer_rename_cb)
        self.__update_name()
        name_row.pack_start(self.name_entry, True, True, 0)

        self.audio_button = Gtk.Button.new()
        self.audio_button.connect("clicked", self.__audio_button_clicked_cb)
        self.__update_audio_button()

        self.video_button = Gtk.Button.new()
        self.video_button.connect("clicked", self.__video_button_clicked_cb)
        self.__update_video_button()

        control_box = Gtk.ButtonBox()
        control_box.set_layout(Gtk.ButtonBoxStyle.EXPAND)
        control_box.add(self.video_button)
        control_box.add(self.audio_button)
        name_row.pack_start(control_box, False, False, 0)

        space = Gtk.Label()
        space.props.vexpand = True
        vbox.pack_start(space, False, False, 0)

        self.ges_layer.connect("notify::priority", self.__layer_priority_changed_cb)
        self.ges_layer.connect("active-changed", self.__layer_active_changed_cb)
        self.ges_timeline.connect("layer-added", self.__timeline_layer_added_cb)
        self.ges_timeline.connect("layer-removed", self.__timeline_layer_removed_cb)
        self.__update_actions()

        # When the window property is set, specify the mouse cursor.
        self.connect("notify::window", self.__window_set_cb)

    def __window_set_cb(self, unused_window, unused_pspec):
        self.props.window.set_cursor(Gdk.Cursor.new(Gdk.CursorType.HAND1))

    def __del__(self):
        self.name_entry.disconnect_by_func(self.__name_focus_out_cb)
        self.ges_layer.disconnect_by_func(self.__layer_rename_cb)
        self.ges_layer.disconnect_by_func(self.__layer_priority_changed_cb)
        self.ges_timeline.disconnect_by_func(self.__timeline_layer_added_cb)
        self.ges_timeline.disconnect_by_func(self.__timeline_layer_removed_cb)
        super().__del__()

    def __layer_rename_cb(self, unused_ges_layer, item, value):
        if not item == "video::name":
            return
        self.__update_name()

    def __name_focus_out_cb(self, unused_widget, unused_event):
        self.name_entry.delete_selection()
        current_name = self.ges_layer.ui.get_name()
        name = self.name_entry.get_text()
        if name == current_name:
            return

        with self.app.action_log.started("change layer name",
                                         toplevel=True):
            self.ges_layer.ui.set_name(name)

    def __layer_priority_changed_cb(self, unused_ges_layer, unused_pspec):
        self.__update_actions()
        self.__update_name()

    def __timeline_layer_added_cb(self, unused_timeline, unused_ges_layer):
        self.__update_actions()

    def __timeline_layer_removed_cb(self, unused_timeline, unused_ges_layer):
        self.__update_actions()

    def __update_actions(self):
        priority = self.ges_layer.get_priority()
        first = priority == 0
        self.__move_layer_up_action.props.enabled = not first
        self.__move_layer_top_action.props.enabled = not first
        layers_count = len(self.ges_timeline.get_layers())
        last = priority == layers_count - 1
        self.__move_layer_down_action.props.enabled = not last
        self.__move_layer_bottom_action.props.enabled = not last
        self.delete_layer_action.props.enabled = layers_count > 1

    def __update_name(self):
        self.name_entry.set_text(self.ges_layer.ui.get_name())

    def __create_menu_model(self):
        action_group = Gio.SimpleActionGroup()
        menu_model = Gio.Menu()

        self.__move_layer_top_action = Gio.SimpleAction.new("move-layer-to-top", None)
        action = self.__move_layer_top_action
        action.connect("activate", self.__move_layer_cb, -2)
        action_group.add_action(action)
        menu_model.append(_("Move layer to top"), "layer.%s" % action.get_name().replace(" ", "."))

        self.__move_layer_up_action = Gio.SimpleAction.new("move-layer-up", None)
        action = self.__move_layer_up_action
        action.connect("activate", self.__move_layer_cb, -1)
        action_group.add_action(action)
        menu_model.append(_("Move layer up"), "layer.%s" % action.get_name().replace(" ", "."))

        self.__move_layer_down_action = Gio.SimpleAction.new("move-layer-down", None)
        action = self.__move_layer_down_action
        action.connect("activate", self.__move_layer_cb, 1)
        action_group.add_action(action)
        menu_model.append(_("Move layer down"), "layer.%s" % action.get_name().replace(" ", "."))

        self.__move_layer_bottom_action = Gio.SimpleAction.new("move-layer-to-bottom", None)
        action = self.__move_layer_bottom_action
        action.connect("activate", self.__move_layer_cb, 2)
        action_group.add_action(action)
        menu_model.append(_("Move layer to bottom"), "layer.%s" % action.get_name().replace(" ", "."))

        self.delete_layer_action = Gio.SimpleAction.new("delete-layer", None)
        action = self.delete_layer_action
        action.connect("activate", self.__delete_layer_cb)
        action_group.add_action(action)
        menu_model.append(_("Delete layer"), "layer.%s" % action.get_name())

        return menu_model, action_group

    def __delete_layer_cb(self, unused_action, unused_parameter):
        pipeline = self.ges_timeline.get_asset().pipeline
        with self.app.action_log.started("delete layer",
                                         finalizing_action=CommitTimelineFinalizingAction(pipeline),
                                         toplevel=True):
            self.ges_timeline.remove_layer(self.ges_layer)
            removed_priority = self.ges_layer.props.priority
            for ges_layer in self.ges_timeline.get_layers():
                if ges_layer.props.priority > removed_priority:
                    ges_layer.props.priority -= 1

    def __move_layer_cb(self, unused_simple_action, unused_parameter, step):
        index = self.ges_layer.get_priority()
        if abs(step) == 1:
            index += step
        elif step == -2:
            index = 0
        else:
            index = len(self.ges_timeline.get_layers()) - 1
        self.ges_timeline.ui.move_layer(self.ges_layer, index)
        self.app.project_manager.current_project.pipeline.commit_timeline()

    def __audio_button_clicked_cb(self, button):
        self.ges_layer.set_active_for_tracks(not self.__check_tracks_active(
            self.timeline_audio_tracks), self.timeline_audio_tracks)
        self.app.project_manager.current_project.pipeline.commit_timeline()

    def __update_audio_button(self):
        active = self.__check_tracks_active(self.timeline_audio_tracks)
        icon = AUDIO_ICONS[active]
        self.audio_button.set_image(Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON))

    def __video_button_clicked_cb(self, button):
        self.ges_layer.set_active_for_tracks(not self.__check_tracks_active(
            self.timeline_video_tracks), self.timeline_video_tracks)
        self.app.project_manager.current_project.pipeline.commit_timeline()

    def __update_video_button(self):
        active = self.__check_tracks_active(self.timeline_video_tracks)
        icon = VIDEO_ICONS[active]
        self.video_button.set_image(Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON))

    def __layer_active_changed_cb(self, ges_layer, active, tracks):
        self.__update_video_button()
        self.__update_audio_button()

    def __check_tracks_active(self, tracks):
        return all(self.ges_layer.get_active_for_track(t) for t in tracks)

    def update(self, media_types: GES.TrackType):
        self.props.height_request = self.ges_layer.ui.props.height_request

        has_audio = media_types & GES.TrackType.AUDIO
        self.audio_button.set_sensitive(has_audio)

        has_video = media_types & GES.TrackType.VIDEO
        self.video_button.set_sensitive(has_video)

        if has_video or not media_types:
            # The layer has video or is empty.
            icon = "video-x-generic-symbolic"
        else:
            # The layer has audio and nothing else.
            icon = "audio-x-generic-symbolic"

        if icon != self.__icon:
            image = Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON)
            self.menubutton.props.image = image
            self.__icon = icon


class Layer(Gtk.Layout, Zoomable, Loggable):
    """Container for the clips widgets of a layer."""

    __gtype_name__ = "PitiviLayer"

    def __init__(self, ges_layer, timeline):
        Gtk.Layout.__init__(self)
        Zoomable.__init__(self)
        Loggable.__init__(self)

        self.ges_layer = ges_layer
        self.ges_layer.ui = self
        self.timeline = timeline
        self.app = timeline.app

        self._children = []
        self._changed = False

        self.ges_layer.connect("clip-added", self._clip_added_cb)
        self.ges_layer.connect("clip-removed", self._clip_removed_cb)

        # The layer is always the width of the Timeline which contains it.
        self.props.hexpand = True
        self.props.valign = Gtk.Align.START

        self.media_types = GES.TrackType(0)
        for ges_clip in ges_layer.get_clips():
            self._add_clip(ges_clip)
        self.check_media_types()

    def set_name(self, name):
        self.ges_layer.set_meta("video::name", name)

    def _name_if_set(self):
        name = self.ges_layer.get_meta("video::name")
        if not name:
            name = self.ges_layer.get_meta("audio::name")
        return name

    def __name_if_meaningful(self):
        name = self._name_if_set()
        if name:
            for pattern in ("video [0-9]+$", "audio [0-9]+$", "Layer [0-9]+$"):
                if re.match(pattern, name):
                    return None
        return name

    def get_name(self):
        name = self.__name_if_meaningful()
        if not name:
            name = _('Layer %d') % self.ges_layer.get_priority()
        return name

    def release(self):
        for ges_clip in self.ges_layer.get_clips():
            self._remove_clip(ges_clip)
        self.ges_layer.disconnect_by_func(self._clip_added_cb)
        self.ges_layer.disconnect_by_func(self._clip_removed_cb)

    def check_media_types(self):
        if self.timeline.editing_context:
            self.info("Not updating media types as we are editing the timeline")
            return

        old_media_types = self.media_types
        self.media_types: GES.TrackType = GES.TrackType(0)
        ges_clips = self.ges_layer.get_clips()
        for ges_clip in ges_clips:
            self.media_types |= ges_clip.props.supported_formats
            if self.media_types & GES.TrackType.AUDIO and self.media_types & GES.TrackType.VIDEO:
                # Cannot find more types than these.
                break

        if self.media_types & GES.TrackType.AUDIO and self.media_types & GES.TrackType.VIDEO:
            self.props.height_request = LAYER_HEIGHT
        else:
            # If the layer is empty, set layer's height to default height.
            self.props.height_request = LAYER_HEIGHT // 2

        if hasattr(self.ges_layer, "control_ui") and self.ges_layer.control_ui:
            self.ges_layer.control_ui.update(self.media_types)

        if old_media_types != self.media_types:
            self.update_position()

    def _clip_child_added_cb(self, ges_clip, child):
        self.check_media_types()

    def _clip_child_removed_cb(self, ges_clip, child):
        self.check_media_types()

    def _clip_added_cb(self, unused_ges_layer, ges_clip):
        self._add_clip(ges_clip)
        self.check_media_types()

    def _add_clip(self, ges_clip):
        ui_type = elements.GES_TYPE_UI_TYPE.get(ges_clip.__gtype__, None)
        if ui_type is None:
            self.error("Implement UI for type %s?", ges_clip.__gtype__)
            return

        widget = ui_type(self, ges_clip)
        self._children.append(widget)
        self._children.sort(key=lambda clip: clip.z_order)
        self.put(widget, self.ns_to_pixel(ges_clip.props.start), 0)
        widget.update_position()
        self._changed = True
        widget.show_all()

        ges_clip.connect_after("child-added", self._clip_child_added_cb)
        ges_clip.connect_after("child-removed", self._clip_child_removed_cb)

    def _clip_removed_cb(self, unused_ges_layer, ges_clip):
        self._remove_clip(ges_clip)
        self.check_media_types()

    def _remove_clip(self, ges_clip):
        if not ges_clip.ui:
            return

        ui_type = elements.GES_TYPE_UI_TYPE.get(ges_clip.__gtype__, None)
        if ui_type is None:
            self.error("Implement UI for type %s?", ges_clip.__gtype__)
            return

        self.remove(ges_clip.ui)
        self._children.remove(ges_clip.ui)
        self._changed = True
        ges_clip.ui.release()
        ges_clip.ui = None

        ges_clip.disconnect_by_func(self._clip_child_added_cb)
        ges_clip.disconnect_by_func(self._clip_child_removed_cb)

    def update_position(self):
        for ges_clip in self.ges_layer.get_clips():
            if hasattr(ges_clip, "ui"):
                ges_clip.ui.update_position()

    def do_draw(self, cr):
        if self._changed:
            self._children.sort(key=lambda clip: clip.z_order)
            for child in self._children:
                if isinstance(child, elements.TransitionClip):
                    window = child.get_window()
                    window.raise_()
            self._changed = False

        for child in self._children:
            self.propagate_draw(child, cr)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Ayush Mittal <ayush.mittal9398@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import Gtk

from pitivi.utils.ui import clear_styles
from pitivi.utils.ui import gtk_style_context_get_color
from pitivi.utils.ui import PADDING


class ActionSearchBar(Gtk.Window):

    def __init__(self, app):
        Gtk.Window.__init__(self)
        self.app = app

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        self.set_size_request(600, 50)
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=PADDING)
        self.set_decorated(False)
        self.add(self.vbox)

        self.entry = Gtk.SearchEntry()
        self.results_window = Gtk.ScrolledWindow()
        self.results_window.props.can_focus = False
        clear_styles(self.entry)
        self.setup_results_window()
        self.entry.props.placeholder_text = _("Search Action")

        self.entry.connect("key-press-event", self._entry_key_press_event_cb)
        self.entry.connect("changed", self._entry_changed_cb)
        self.treeview.connect("row-activated", self._treeview_row_activated_cb)

        self.vbox.pack_start(self.entry, True, True, 0)
        self.vbox.pack_start(self.results_window, True, True, 0)

    def setup_results_window(self):
        self.list_model = Gtk.ListStore(str, int, Gdk.ModifierType, Gio.SimpleAction, object, bool, bool)
        self.model_filter = self.list_model.filter_new()
        disable_groups = ["medialibrary"]

        style_context = self.entry.get_style_context()
        color_insensitive = gtk_style_context_get_color(style_context, Gtk.StateFlags.INSENSITIVE)

        for group in self.app.shortcuts.group_actions:
            if group not in disable_groups:
                for action, title, action_object in self.app.shortcuts.group_actions[group]:
                    accels = self.app.get_accels_for_action(action)
                    accel = accels[0] if accels else ""
                    accelerator_parsed = Gtk.accelerator_parse(accel)
                    disabled = not action_object.props.enabled
                    self.list_model.append([title,
                                            accelerator_parsed.accelerator_key,
                                            accelerator_parsed.accelerator_mods,
                                            action_object,
                                            title.lower().split(" "),
                                            disabled,
                                            bool(accels)])

        self.model_filter.set_visible_func(self.filter_func)
        self.treeview = Gtk.TreeView.new_with_model(self.model_filter)
        self.treeview.props.headers_visible = False
        self.treeview.props.enable_search = False
        self.treeview.props.can_focus = False

        text_renderer = Gtk.CellRendererText()
        text_renderer.props.foreground_rgba = color_insensitive
        description_column = Gtk.TreeViewColumn("Description", text_renderer, text=0, foreground_set=5)
        description_column.set_fixed_width(400)
        self.treeview.append_column(description_column)

        accel_renderer = Gtk.CellRendererAccel()
        # The default is Gtk.CellRendererAccelMode.GTK, but with that one
        # accelerator "Left" appears as "Invalid" for some reason.
        accel_renderer.props.accel_mode = Gtk.CellRendererAccelMode.OTHER
        accel_renderer.props.foreground_rgba = color_insensitive
        accel_renderer.props.foreground_set = True
        shortcut_column = Gtk.TreeViewColumn("Shortcut", accel_renderer, accel_key=1, accel_mods=2, visible=6)
        self.treeview.append_column(shortcut_column)

        self.__select_row(self.model_filter.get_iter_first())

        self.results_window.add(self.treeview)
        self.results_window.props.min_content_height = 300

    def filter_func(self, model, row, data):
        text = self.entry.get_text()
        if text:
            self.results_window.show()
            all_found = True
            for keyword in text.lower().split(" "):
                found = False
                for term in model[row][4]:
                    if term.startswith(keyword):
                        found = True
                        break
                if not found:
                    all_found = False
            return all_found
        else:
            return True

    def do_focus_out_event(self, event):
        self.destroy()
        return True

    def _treeview_row_activated_cb(self, treeview, path, col):
        self.__activate_selected_action()

    def __activate_selected_action(self):
        model, row_iter = self.treeview.get_selection().get_selected()
        if not row_iter:
            # No row is selected, possibly because there are no rows.
            return

        action_object, = model.get(row_iter, 3)
        action_object.activate()
        self.destroy()

    def _entry_changed_cb(self, entry):
        self.model_filter.refilter()

        # Make sure a row is always selected.
        self.__select_row(self.model_filter.get_iter_first())

    def __select_row(self, row_iter):
        if not row_iter:
            return

        self.treeview.get_selection().select_iter(row_iter)

        row_path = self.model_filter.get_path(row_iter)
        self.treeview.scroll_to_cell(row_path, None, False, 0, 0)

    def _entry_key_press_event_cb(self, entry, event):
        if event.keyval == Gdk.KEY_Escape:
            self.destroy()
            return True

        if event.keyval == Gdk.KEY_Return:
            self.__activate_selected_action()
            return True

        if event.keyval == Gdk.KEY_Up:
            selection = self.treeview.get_selection()
            model, row_iter = selection.get_selected()
            if row_iter:
                self.__select_row(model.iter_previous(row_iter))
                return True

        if event.keyval == Gdk.KEY_Down:
            selection = self.treeview.get_selection()
            model, row_iter = selection.get_selected()
            if row_iter:
                self.__select_row(model.iter_next(row_iter))
                return True

        # Let the default handler process this event.
        return False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Pratyush Tiwari <pratyushtiwarimj@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk


INTERACTIVE_INTRO_CSS = """
@keyframes intro-highlight {
  from { box-shadow: 0px 0px 10px @theme_selected_bg_color, inset 0px 0px 4px  @theme_selected_bg_color; }
  to   { box-shadow: 0px 0px 4px  @theme_selected_bg_color, inset 0px 0px 10px @theme_selected_bg_color; }
}

.intro-highlight {
  animation: intro-highlight 1.5s infinite alternate;
}


#control-popover > box > button {
  margin: 10px;
}


#intro-popover {
  background-color: @theme_selected_bg_color;
}

#intro-popover > box {
  margin: 10px;
}

#intro-popover > box > label {
  font-size: 18pt;
  font-weight: bold;
  margin: 10px;
  color: @theme_fg_color;
}
"""


class InteractiveIntro(GObject.Object):
    """Interactive GUI intro for newcomers.

    Attributes:
        intro_action (Gio.SimpleAction): The action for showing the control
            popover.
        intro_button (Gtk.Button): The button to be displayed in the headerbar
            for showing the control popover.
        widget (Gtk.Widget): The current widget being highlighted.
    """

    def __init__(self, app):
        GObject.Object.__init__(self)
        self.app = app
        self.widget = None
        self.parent_widget = None
        self.tips = []
        self._advance_handler_id = 0
        self._hiding_popover_when_advancing = False

        self.__setup_css()

        self.intro_action = Gio.SimpleAction.new("interactive-intro", None)
        self.intro_action.connect("activate", self._control_intro_cb)

        self.intro_button = self.__create_intro_button()

    def __setup_css(self):
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(INTERACTIVE_INTRO_CSS.encode("UTF-8"))
        screen = Gdk.Screen.get_default()
        style_context = self.app.gui.get_style_context()
        style_context.add_provider_for_screen(screen, css_provider,
                                              Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def __create_intro_button(self):
        """Creates a button for controlling the intro."""
        button = Gtk.Button.new_from_icon_name(
            "org.pitivi.Pitivi-symbolic", Gtk.IconSize.LARGE_TOOLBAR)
        button.set_always_show_image(True)
        button.props.no_show_all = True
        button.props.relief = Gtk.ReliefStyle.NONE
        # We could set_action_name, but we don't know here the group
        # in which the action will be added, so it's simpler this way.
        button.connect("clicked", self._intro_button_clicked_cb)
        return button

    def __create_popover(self, text):
        popover = Gtk.Popover()
        popover.set_position(Gtk.PositionType.BOTTOM)
        popover.set_modal(True)
        popover.connect("closed", self._intro_popover_closed_cb)
        popover.set_name("intro-popover")

        label = Gtk.Label()
        label.set_markup(text)
        label.set_line_wrap(True)
        label.set_max_width_chars(24)

        img = Gtk.Image.new_from_icon_name("dialog-information-symbolic", Gtk.IconSize.DIALOG)

        vbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        vbox.pack_start(img, False, False, 5)
        vbox.pack_end(label, False, False, 5)
        vbox.show_all()

        popover.add(vbox)

        return popover

    def _intro_popover_closed_cb(self, unused_widget):
        if self._hiding_popover_when_advancing:
            if self.parent_widget:
                self.parent_widget.hide()
                self.parent_widget = None
            return

        self.stop_tour()

    def _control_intro_cb(self, unused_action, unused_param):
        """Handles the activation of the interactive intro action."""
        if self.running:
            self.stop_tour()
        else:
            self.show_control_popover()

    def _intro_button_clicked_cb(self, button):
        self.intro_action.activate()

    @property
    def running(self):
        return self._advance_handler_id != 0

    def show_control_popover(self):
        # pylint: disable=attribute-defined-outside-init
        overview_button = Gtk.Button.new_with_label(_("Start Overview"))
        overview_button.connect("clicked", self._overview_button_clicked_cb)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(overview_button, False, True, 10)

        self.intro_button.show()
        self.intro_button.get_style_context().add_class("intro-highlight")

        self.control_popover = Gtk.Popover()
        self.control_popover.set_name("control-popover")
        self.control_popover.add(vbox)
        self.control_popover.set_relative_to(self.intro_button)
        self.control_popover.show_all()
        self.control_popover.popup()
        self.control_popover.connect("closed", self._control_popover_closed_cb)

    def _control_popover_closed_cb(self, unused_widget):
        self.intro_button.get_style_context().remove_class("intro-highlight")

    def set_current_widget(self, widget):
        """Switches the focus to a new widget, highlighting it."""
        if self.widget:
            self.widget.get_style_context().remove_class("intro-highlight")
            self.widget = None

        if widget:
            self.widget = widget
            self.widget.get_style_context().add_class("intro-highlight")

            if isinstance(self.widget, Gtk.ModelButton):
                self.parent_widget = self._find_parent(self.widget, Gtk.Popover)
                if self.parent_widget:
                    self.parent_widget.popup()

    def _find_parent(self, widget, parent_class):
        while widget:
            if isinstance(widget, parent_class):
                return widget

            widget = widget.props.parent

        return None

    def stop_tour(self):
        if self._advance_handler_id:
            GLib.source_remove(self._advance_handler_id)
            self._advance_handler_id = 0

        self.set_current_widget(None)

        self.intro_button.hide()

    def _overview_button_clicked_cb(self, unused_widget):
        """Starts the UI overview."""
        self.control_popover.popdown()
        editor = self.app.gui.editor
        self.tips = [
            (self.intro_button, _("Welcome to Pitivi!"),
             Gtk.PositionType.BOTTOM, False, 4000),
            (editor.medialibrary.scrollwin, _("Drag files here to import them"),
             Gtk.PositionType.RIGHT, True, 5000),
            (editor.timeline_ui.timeline, _("Drag clips in the timeline to arrange them"),
             Gtk.PositionType.TOP, True, 7000),
            (editor.timeline_ui.ruler, _("Right-click anywhere to move the playhead"),
             Gtk.PositionType.TOP, True, 6000),
            (editor.viewer.playpause_button, _("Preview with the play button"),
             Gtk.PositionType.TOP, False, 5000),
            (editor.render_button, _("Export your project to share it"),
             Gtk.PositionType.BOTTOM, False, 5000),
            (editor.keyboard_shortcuts_button, _("See all the keyboard shortcuts"),
             Gtk.PositionType.LEFT, False, 6000)]

        self.advance_popover_func(previous_popover=None)

    def advance_popover_func(self, previous_popover):
        """Moves the popover to the next widget."""
        if previous_popover:
            self._hiding_popover_when_advancing = True
            try:
                previous_popover.popdown()
            finally:
                self._hiding_popover_when_advancing = False

        if not self.tips:
            self.stop_tour()
            return False

        widget, text, position_type, center, timeout = self.tips.pop(0)
        self.set_current_widget(widget)

        popover = self.__create_popover(text)
        popover.set_relative_to(widget)

        if center:
            rect = Gdk.Rectangle()
            rect.x = widget.get_allocated_width() / 2
            rect.y = widget.get_allocated_height() / 2
            rect.width = 4
            rect.height = 4
            popover.set_pointing_to(rect)

        popover.set_position(position_type)
        popover.popup()

        self._advance_handler_id = GLib.timeout_add(timeout, self.advance_popover_func, popover)
        return False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Thibault Saunier <tsaunier@igalia.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Simple tool used by Pitivi to check if a GStreamer pipeline can run."""
import os
import sys

import gi

gi.require_version("Gst", "1.0")

# pylint: disable-msg=wrong-import-position
from gi.repository import Gst
from gi.repository import GLib


def pipeline_message_cb(_, msg, pipeline):
    """Handles a GStreamer bus message."""
    if msg.type == Gst.MessageType.ASYNC_DONE:
        print("pipeline successfully PAUSED")
        pipeline.set_state(Gst.State.NULL)
        sys.exit(0)
    elif msg.type == Gst.MessageType.ERROR:
        # The pipeline cannot be set to PAUSED.
        error, detail = msg.parse_error()
        print("check_pipeline: Pipeline failed: %s, %s" % (error, detail), file=sys.stderr)
        pipeline.set_state(Gst.State.NULL)
        sys.exit(1)


def timeout_cb(*args, **kwargs):
    """Exits on timeout."""
    print("check_pipeline: Pipeline timed out", file=sys.stderr)
    sys.exit(1)


def main():
    """Runs this small tool."""
    os.environ["G_DEBUG"] = "fatal-criticals"
    Gst.init(None)
    pipeline = Gst.parse_launch(sys.argv[1])
    bus = pipeline.get_bus()
    bus.add_signal_watch()
    bus.connect("message", pipeline_message_cb, pipeline)
    res = pipeline.set_state(Gst.State.PAUSED)
    assert res == Gst.StateChangeReturn.ASYNC

    GLib.timeout_add_seconds(10, timeout_cb)
    GLib.MainLoop.new(None, True).run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Pitivi's main window."""
import os
from gettext import gettext as _
from urllib.parse import unquote

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from pitivi.configure import get_pixmap_dir
from pitivi.dialogs.about import AboutDialog
from pitivi.dialogs.prefs import PreferencesDialog
from pitivi.editorperspective import EditorPerspective
from pitivi.greeterperspective import GreeterPerspective
from pitivi.settings import GlobalSettings
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import show_user_manual


GlobalSettings.add_config_option('mainWindowX',
                                 section="main-window",
                                 key="X", default=0, type_=int)
GlobalSettings.add_config_option('mainWindowY',
                                 section="main-window",
                                 key="Y", default=0, type_=int)
GlobalSettings.add_config_option('mainWindowWidth',
                                 section="main-window",
                                 key="width", default=-1, type_=int)
GlobalSettings.add_config_option('mainWindowHeight',
                                 section="main-window",
                                 key="height", default=-1, type_=int)

GlobalSettings.add_config_section('export')
GlobalSettings.add_config_option('lastExportFolder',
                                 section='export',
                                 key="last-export-folder",
                                 environment="PITIVI_EXPORT_FOLDER",
                                 default=os.path.expanduser("~"))

GlobalSettings.add_config_section("version")
GlobalSettings.add_config_option('displayCounter',
                                 section='version',
                                 key='info-displayed-counter',
                                 default=0)
GlobalSettings.add_config_option('lastCurrentVersion',
                                 section='version',
                                 key='last-current-version',
                                 default='')

GlobalSettings.add_config_option("useDarkTheme",
                                 section="user-interface",
                                 key="use-dark-theme",
                                 default=True,
                                 notify=True)

PreferencesDialog.add_toggle_preference("useDarkTheme",
                                        section="other",
                                        label=_("Dark Theme"),
                                        description=_("Whether or not to use a dark theme."))


class MainWindow(Gtk.ApplicationWindow, Loggable):
    """Pitivi's main window.

    It manages the UI and handles the switch between different perspectives,
    such as the default GreeterPerspective, and the EditorPerspective.

    Attributes:
        app (Pitivi): The app.
    """

    def __init__(self, app):
        # Pulseaudio "role"
        # (http://0pointer.de/blog/projects/tagging-audio.htm)
        os.environ["PULSE_PROP_media.role"] = "production"
        os.environ["PULSE_PROP_application.icon_name"] = "pitivi"

        Gtk.IconTheme.get_default().append_search_path(get_pixmap_dir())
        Gtk.IconTheme.get_default().append_search_path(os.path.join(get_pixmap_dir(), "transitions"))

        Gtk.ApplicationWindow.__init__(self)
        Loggable.__init__(self)

        self.log("Creating main window")

        self.app = app
        self.greeter = GreeterPerspective(app)
        self.editor = EditorPerspective(app)
        self.__placed = False
        self.__perspective = None
        self.__wanted_perspective = None

        self.app.settings.connect("useDarkThemeChanged",
                                  self.__use_dark_theme_changed_cb)
        self.update_use_dark_theme()

        app.project_manager.connect("new-project-loading",
                                    self.__new_project_loading_cb)
        app.project_manager.connect("new-project-failed",
                                    self.__new_project_failed_cb)
        app.project_manager.connect("project-closed", self.__project_closed_cb)

    def update_use_dark_theme(self):
        gtksettings = Gtk.Settings.get_default()
        use_dark_theme = self.app.settings.useDarkTheme
        gtksettings.set_property("gtk-application-prefer-dark-theme", use_dark_theme)

    def __use_dark_theme_changed_cb(self, unused_settings):
        self.update_use_dark_theme()

    def setup_ui(self):
        """Sets up the various perspectives's UI."""
        self.log("Setting up the perspectives.")

        self.set_icon_name("org.pitivi.Pitivi")
        self._create_actions()

        self.greeter.setup_ui()
        self.editor.setup_ui()

        width = self.app.settings.mainWindowWidth
        height = self.app.settings.mainWindowHeight
        if height == -1 and width == -1:
            self.__placed = True
            self.maximize()
        else:
            # Wait until placing the window, to avoid the window manager
            # ignoring the call. See the documentation for Gtk.Window.move.
            # If you change this, pay attention opening `pitivi` works
            # a bit different than opening `pitivi file.xges`. For example
            # connecting to the "realize" signal instead of idle_add-ing
            # fails to restore the position when directly loading a project.
            GLib.idle_add(self.__initial_placement_cb,
                          self.app.settings.mainWindowX,
                          self.app.settings.mainWindowY,
                          width, height)

        self.check_screen_constraints()

        self.connect("configure-event", self.__configure_cb)
        self.connect("delete-event", self.__delete_cb)

    def __initial_placement_cb(self, x, y, width, height):
        self.__placed = True
        self.resize(width, height)
        self.move(x, y)
        if self.__wanted_perspective:
            self.show_perspective(self.__wanted_perspective)
            self.__wanted_perspective = None

    def check_screen_constraints(self):
        """Shrinks some widgets to fit better on smaller screen resolutions."""
        if self._small_screen():
            self.greeter.activate_compact_mode()
            self.editor.activate_compact_mode()
            min_size, _ = self.get_preferred_size()
            self.info("Minimum UI size has been reduced to %sx%s",
                      min_size.width, min_size.height)

    def _small_screen(self):
        # This code works, but keep in mind get_preferred_size's output
        # is only an approximation. As of 2015, GTK still does not have
        # a way, even with client-side decorations, to tell us the exact
        # minimum required dimensions of a window.
        min_size, _ = self.get_preferred_size()
        screen_width = self.get_screen().get_width()
        screen_height = self.get_screen().get_height()
        self.debug("Minimum UI size is %sx%s", min_size.width, min_size.height)
        self.debug("Screen size is %sx%s", screen_width, screen_height)
        return min_size.width >= 0.9 * screen_width

    def _create_actions(self):
        self.app.shortcuts.register_group("win", _("Project"), position=20)

        # pylint: disable=attribute-defined-outside-init
        self.help_action = Gio.SimpleAction.new("help", None)
        self.help_action.connect("activate", self.__user_manual_cb)
        self.add_action(self.help_action)
        self.app.shortcuts.add("win.help", ["F1"], self.help_action,
                               _("Help"), group="app")

        self.about_action = Gio.SimpleAction.new("about", None)
        self.about_action.connect("activate", self.__about_cb)
        self.add_action(self.about_action)
        self.app.shortcuts.add("win.about", ["<Primary><Shift>a"],
                               self.about_action,
                               _("About"), group="app")

        self.main_menu_action = Gio.SimpleAction.new("menu-button", None)
        self.main_menu_action.connect("activate", self.__menu_cb)
        self.add_action(self.main_menu_action)
        self.app.shortcuts.add("win.menu-button", ["F10"],
                               self.main_menu_action,
                               _("Show the menu button content"), group="app")

        self.preferences_action = Gio.SimpleAction.new("preferences", None)
        self.preferences_action.connect("activate", self.__preferences_cb)
        self.add_action(self.preferences_action)
        self.app.shortcuts.add("win.preferences", ["<Primary>comma"],
                               self.preferences_action,
                               _("Preferences"), group="app")

    @staticmethod
    def __user_manual_cb(unused_action, unused_param):
        show_user_manual()

    def __about_cb(self, unused_action, unused_param):
        about_dialog = AboutDialog(self.app)
        about_dialog.show()

    def __menu_cb(self, unused_action, unused_param):
        active = not self.__perspective.menu_button.get_active()
        self.__perspective.menu_button.set_active(active)

    def __preferences_cb(self, unused_action, unused_param):
        PreferencesDialog(self.app).run()

    def __configure_cb(self, unused_widget, unused_event):
        """Saves the main window position and size."""
        position = self.get_position()
        self.app.settings.mainWindowX = position.root_x
        self.app.settings.mainWindowY = position.root_y

        size = self.get_size()
        self.app.settings.mainWindowWidth = size.width
        self.app.settings.mainWindowHeight = size.height

    def __delete_cb(self, unused_widget, unused_event):
        self.app.settings.mainWindowHPanePosition = self.editor.secondhpaned.get_position()
        self.app.settings.mainWindowMainHPanePosition = self.editor.mainhpaned.get_position()
        self.app.settings.mainWindowVPanePosition = self.editor.toplevel_widget.get_position()

        if not self.app.shutdown():
            return True
        return False

    def __new_project_loading_cb(self, unused_project_manager, unused_project):
        self.show_perspective(self.editor)

    def __new_project_failed_cb(self, unused_project_manager, uri, reason):
        project_filename = unquote(uri.split("/")[-1])
        dialog = Gtk.MessageDialog(transient_for=self,
                                   modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=_('Unable to load project "%s"') % project_filename)
        dialog.set_property("secondary-use-markup", True)
        dialog.set_property("secondary-text", unquote(str(reason)))
        dialog.set_transient_for(self)
        dialog.run()
        dialog.destroy()
        self.show_perspective(self.greeter)

    def __project_closed_cb(self, unused_project_manager, unused_project):
        self.show_perspective(self.greeter)

    def show_perspective(self, perspective):
        """Displays the specified perspective."""
        if not self.__placed:
            self.log("Postponing the perspective setting until the window is placed")
            self.__wanted_perspective = perspective
            return

        if self.__perspective is perspective:
            return

        if self.__perspective:
            # Remove the current perspective before adding the
            # specified perspective because we can only add one
            # toplevel widget to the main window at a time.
            self.remove(self.__perspective.toplevel_widget)
        self.log("Displaying perspective: %s", type(perspective).__name__)
        self.__perspective = perspective
        self.set_titlebar(perspective.headerbar)
        # The window must be shown only after setting the headerbar with
        # set_titlebar. Otherwise we get a warning things can go wrong.
        self.show()
        self.add(perspective.toplevel_widget)
        perspective.refresh()
